package de.renew.gui;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import collections.CollectionEnumeration;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import de.renew.simulator.*;
import de.renew.event.*;
import de.renew.unify.Variable;

class TransitionInstanceFigure extends AbstractFigure
    implements InstanceFigure,
	       TransitionEventListener {

  private TransitionFigure transitionFigure;
  private TransitionInstance transitionInstance;
  private int actionCounter=0;
  private boolean afterglow=false;
  private CPNInstanceDrawing drawing;

  public TransitionInstanceFigure(CPNInstanceDrawing drawing,
				  TransitionFigure tf, TransitionInstance ti) {
	this.drawing = drawing;

        transitionInstance = ti;
        transitionFigure = tf;

	synchronized (transitionInstance) {
	  CollectionEnumeration enum=transitionInstance.pendingFiringEvents();
	  while (enum.hasMoreElements()) {
	    FiringEvent fe=(FiringEvent)enum.nextElement();
	    firingStarted(fe);
	  }
	}

	transitionInstance.addTransitionEventListener(this);
  }

  public boolean isHighlighted() {
      return actionCounter > 0 || afterglow;
  }

  private void invalidateHighlight() {
      Figure highlight=transitionFigure.getHighlightFigure();
      if (highlight!=null) {
	  Rectangle area = highlight.displayBox();
	  area.grow(5,5);
	  DrawingChangeEvent dce = new DrawingChangeEvent(drawing,area);
	  drawing.drawingInvalidated(dce);
      }
  }


  public boolean wantSynchronousNotification() {
    return false;
  }

    /** This event occurs when a TransitionInstance starts a new
     *  Occurrence. When the event is sent, it is guaranteed
     *  that the firing can take place, i.e., all input tokens have 
     *  already been successfully collected.
     *
     *  The Event is sent to all TransitionInstances which are part of the
     *  Occurrence, not only to the triggering TransitionInstance.
     */
    public void firingStarted(FiringEvent fe) {
      try {
	drawing.lock();
	
	//System.out.println("Firing "+transitionInstance);
	++actionCounter;
	if (actionCounter == 1 && listener() != null) {
	  invalidate();
	  invalidateHighlight();
	  listener().figureRequestUpdate(new FigureChangeEvent(this));
	}
      } finally {
	drawing.unlock();
      }
    }

    /** This event signals that a Transition Occurrence has completed.
     *  The Event is sent to all TransitionInstances which are part of the
     *  Occurrence, not only to the triggering TransitionInstance.
     *
     *  The event that is delivered is exactly the same event that was
     *  delivered during the start of the firing.
     */
    public void firingComplete(FiringEvent fe) {
      try {
	drawing.lock();
	
	//System.out.println("Firing of "+transitionInstance+" completed.");
	--actionCounter;
	if (actionCounter == 0 && listener() != null) {
	  afterglow=true;
	  final TransitionInstanceFigure thizz=this;
	  new Thread() {
	      public void run() {
		try { sleep(100); }
		catch (InterruptedException e) { }
		try {
		  drawing.lock();
		  afterglow=false;
		  invalidate();
		  invalidateHighlight();
		  listener().figureRequestUpdate(new FigureChangeEvent(thizz));
		} finally {
		  drawing.unlock();
		}
	      }
	    }.start();
	}
      } finally {
	drawing.unlock();
      }
    }

    public void basicDisplayBox(Point origin, Point corner) {
        // do nothing, as the DisplayBox of the Transition is used!
    }

    public Vector handles() {
        Vector handles = new Vector();
        handles.addElement(new NullHandle(this, RelativeLocator.northWest()));
        handles.addElement(new NullHandle(this, RelativeLocator.northEast()));
        handles.addElement(new NullHandle(this, RelativeLocator.southWest()));
        handles.addElement(new NullHandle(this, RelativeLocator.southEast()));
        return handles;
    }

    public Rectangle displayBox() {
        return transitionFigure.displayBox();
    }

    protected void basicMoveBy(int x, int y) {
        // do nothing, as the DisplayBox of the Transition is used!
    }

    public void draw(Graphics g) {
	/*	if (actionCounter > 0) {
	    g.setColor(Color.red);
	    Rectangle r = displayBox();
	    g.fillRect(r.x, r.y, r.width, r.height);
	}
	*/
    }


  public boolean inspect(DrawingView view, boolean alternate) {
    //System.out.println("Searching bindings for "+transitionInstance+"...");
    if (alternate) {
      if (fire()) {
	view.editor().showStatus("Transition "+transitionInstance+" fired.");
      } else {
	view.editor().showStatus("No binding for this transition.");
      }
    } else {
      BindingSelectionFrame.open(transitionInstance,getSimulation());
    }
    return true;
  }

  private CPNSimulation getSimulation() {
    return drawing.getMode().getSimulation();
  }

  public boolean fire() {
    CPNSimulation simulation=getSimulation();
    simulation.getBreakpointManager().clearLog();
    boolean result=SimulatorHelper.fireOneBinding(transitionInstance,
      new ConsoleTracer("Synchronously","(Automatically) "), 
      !simulation.isStrictlySequential());
    simulation.simulationRefresh();
    return result;
  }

  public void release() {
    if (transitionInstance != null) {
	transitionInstance.removeTransitionEventListener(this);
	// OK: Release reference to transition instance so that it
	// can be garbage collected.
	transitionInstance=null;
    }
    super.release();
  }

  protected TransitionInstance getInstance() {
    return transitionInstance;
  }

  /**
   * This class is not serializable, although inheriting
   * serializability from its superclass.
   * <p>
   * Reason: Serialization is currently used for Copy&Paste
   * and the Undo machanism only. As such modifications don't
   * make sense for instance drawings, this  figure, which is
   * a part of such drawings, doesn't need to be serializable.
   * </p><p>
   * Second reason: Some fields of this class would lead to
   * a serialization of the compiled net, which is not desirable.
   * </p>
   * @throws java.io.NotSerializableException always.
   **/
  private void writeObject(java.io.ObjectOutputStream out)
      throws java.io.IOException {
    throw new java.io.NotSerializableException(
	      "TransitionInstanceFigure is not serializable!");
  }
}
